@@ -2,9 +2,9 @@ package Kossy;
 
 use strict;
 use warnings;
+use 5.008004;
 use utf8;
 use Carp qw//;
-use Scalar::Util qw//;
 use Cwd qw//;
 use File::Basename qw//;
 use Text::Xslate;
@@ -17,10 +17,19 @@ use Class::Accessor::Lite (
     rw => [qw/root_dir/]
 );
 use base qw/Exporter/;
+use Kossy::Exception;
+use Kossy::Connection;
+use Kossy::Request;
+use Kossy::Response;
+use Kossy::Headers;
 
-our $VERSION = '0.24';
+our $VERSION = '0.39';
 our @EXPORT = qw/new root_dir psgi build_app _router _connect get post router filter _wrap_filter/;
 
+our $XSLATE_CACHE = 1;
+our $XSLATE_CACHE_DIR;
+our $SECURITY_HEADER = 1;
+
 sub new {
     my $class = shift;
     my %args;
@@ -62,6 +71,10 @@ sub build_app {
     #router
     my $router = Router::Boom->new;
     $router->add($_ => $self->_router->{$_} ) for keys %{$self->_router};
+    my $xslate_cache_local = $XSLATE_CACHE;
+    my $xslate_cache_dir_local = $XSLATE_CACHE_DIR;
+    my $security_header_local = $SECURITY_HEADER;
+    my %match_cache;
 
     #xslate
     my $fif = HTML::FillInForm::Lite->new();
@@ -78,28 +91,41 @@ sub build_app {
                 }
             }
         },
+        cache => $xslate_cache_local,
+        defined $xslate_cache_dir_local ? ( cache_dir => $xslate_cache_dir_local ) : (),
     );
 
     sub {
         my $env = shift;
+        $Kossy::Headers::SECURITY_HEADER = $security_header_local;
         try {
+            my $header = bless {
+                'content-type' => 'text/html; charset=UTF-8',
+                $security_header_local ? ('x-frame-options' => 'DENY') : (),
+            }, 'Kossy::Headers';
             my $c = Kossy::Connection->new({
                 tx => $tx,
-                req => Kossy::Request->new($env, (parse_json_body => $self->{parse_json_body}) ),
-                res => Kossy::Response->new(200, ['Content-Type' => 'text/html; charset=UTF-8']),
+                req => Kossy::Request->new($env),
+                res => Kossy::Response->new(200, $header),
                 stash => {},
             });
+            my $method = uc($env->{REQUEST_METHOD});
+            my $cache_key = $method . '-' . $env->{PATH_INFO};
             my ($match,$args) = try {
+                if ( exists $match_cache{$cache_key} ) {
+                    return @{$match_cache{$cache_key}};
+                }
                 my $path_info = Encode::decode_utf8( $env->{PATH_INFO},  Encode::FB_CROAK | Encode::LEAVE_SRC );
                 my @match = $router->match($path_info);
                 if ( !@match ) {
                     $c->halt(404);
                 }
-
-                my $method = uc $env->{REQUEST_METHOD};
+                
                 if ( !exists $match[0]->{$method}) {
                     $c->halt(405);
                 }
+                $match_cache{$cache_key} = [$match[0]->{$method},$match[1]]
+                    if ! scalar keys %{$match[1]};
                 return ($match[0]->{$method},$match[1]);
             } catch {
                 if ( ref $_ && ref $_ eq 'Kossy::Exception' ) {
@@ -110,17 +136,18 @@ sub build_app {
             
             my $code = $match->{__action__};
             my $filters = $match->{__filter__} || [];
-            $c->args($args);
+            my %args = map { $_ => Encode::decode_utf8($args->{$_}) } keys %$args;
+            $c->args(\%args);
             my $app = sub {
                 my ($self, $c) = @_;
                 my $response;
                 my $res = $code->($self, $c);
-                Carp::croak "Undefined Response" if !$res;
+                Carp::croak "Undefined Response" if ! defined $res;
                 my $res_t = ref($res) || '';
-                if ( Scalar::Util::blessed $res && $res->isa('Kossy::Response') ) {
-                    $response = $res;;
+                if ( $res_t eq 'Kossy::Response' ) {
+                    $response = $res;
                 }
-                elsif ( Scalar::Util::blessed $res && $res->isa('Plack::Response') ) {
+                elsif ( $res_t eq 'Plack::Response' ) {
                     $response = bless $res, 'Kossy::Response';
                 }
                 elsif ( $res_t eq 'ARRAY' ) {
@@ -217,417 +244,6 @@ sub _wrap_filter {
     return $filter_subref->($app);
 }
 
-package Kossy::Exception;
-
-use strict;
-use warnings;
-use HTTP::Status;
-use Text::Xslate qw/html_escape/;
-
-sub new {
-    my $class = shift;
-    my $code = shift;
-    my %args = (
-        code => $code,
-    );
-    if ( @_ == 1 ) {
-        $args{message} = shift;
-    }
-    elsif ( @_ % 2 == 0) {
-        %args = (
-            %args,
-            @_
-        );
-    }
-    bless \%args, $class;
-}
-
-sub response {
-    my $self = shift;
-    my $code = $self->{code} || 500;
-    my $message = $self->{message};
-    $message ||= HTTP::Status::status_message($code);
-
-    my @headers = (
-         'Content-Type' => q!text/html; charset=UTF-8!,
-    );
-
-    if ($code =~ /^3/ && (my $loc = eval { $self->{location} })) {
-        push(@headers, Location => $loc);
-    }
-
-    return Kossy::Response->new($code, \@headers, [$self->html($code,$message)])->finalize;
-}
-
-sub html {
-    my $self = shift;
-    my ($code,$message) = @_;
-    $code = html_escape($code);
-    $message = html_escape($message);
-    return <<EOF;
-<!doctype html>
-<html>
-<head>
-<meta charset=utf-8 />
-<style type="text/css">
-.message {
-  font-size: 200%;
-  margin: 20px 20px;
-  color: #666;
-}
-.message strong {
-  font-size: 250%;
-  font-weight: bold;
-  color: #333;
-}
-</style>
-</head>
-<body>
-<p class="message">
-<strong>$code</strong> $message
-</p>
-</div>
-</body>
-</html>
-EOF
-}
-
-package Kossy::Connection;
-
-use strict;
-use warnings;
-use Class::Accessor::Lite (
-    new => 1,
-    rw => [qw/req res stash args tx debug/]
-);
-use JSON qw//;
-
-my $_JSON = JSON->new()->allow_blessed(1)->convert_blessed(1)->ascii(1);
-
-# for IE7 JSON venularity.
-# see http://www.atmarkit.co.jp/fcoding/articles/webapp/05/webapp05a.html
-# Copy from Amon2::Plugin::Web::JSON => Fixed to escape only string parts
-my %_ESCAPE = (
-    '+' => '\\u002b', # do not eval as UTF-7
-    '<' => '\\u003c', # do not eval as HTML
-    '>' => '\\u003e', # ditto.
-);
-sub escape {
-    my $self = shift;
-    my $body = shift;
-    $body =~ s!([+<>])!$_ESCAPE{$1}!g;
-    return qq("$body");
-}
-sub escape_json {
-    my $self = shift;
-    my $body = shift;
-    # escape only string parts
-    $body =~ s/"((?:\\"|[^"])*)"/$self->escape($1)/eg;
-    return $body;
-}
-
-*request = \&req;
-*response = \&res;
-
-sub halt {
-    my $self = shift;
-    die Kossy::Exception->new(@_);
-}
-
-sub redirect {
-    my $self = shift;
-    $self->res->redirect(@_);
-    $self->res;
-}
-
-sub render {
-    my $self = shift;
-    my $file = shift;
-    my %args = ( @_ && ref $_[0] ) ? %{$_[0]} : @_;
-    my %vars = (
-        c => $self,
-        stash => $self->stash,
-        %args,
-    );
-
-    my $body = $self->tx->render($file, \%vars);
-    $self->res->status( 200 );
-    $self->res->content_type('text/html; charset=UTF-8');
-    $self->res->body( $body );
-    $self->res;
-}
-
-sub render_json {
-    my $self = shift;
-    my $obj = shift;
-
-    # defense from JSON hijacking
-    # Copy from Amon2::Plugin::Web::JSON
-    if ( !$self->req->header('X-Requested-With')  && 
-         ($self->req->user_agent||'') =~ /android/i &&
-         defined $self->req->header('Cookie') &&
-         ($self->req->method||'GET') eq 'GET'
-    ) {
-        $self->halt(403,"Your request is maybe JSON hijacking.\nIf you are not a attacker, please add 'X-Requested-With' header to each request.");
-    }
-
-    my $body = $_JSON->encode($obj);
-    $body = $self->escape_json($body);
-
-    if ( ( $self->req->user_agent || '' ) =~ m/Safari/ ) {
-        $body = "\xEF\xBB\xBF" . $body;
-    }
-
-    $self->res->status( 200 );
-    $self->res->content_type('application/json; charset=UTF-8');
-    $self->res->header( 'X-Content-Type-Options' => 'nosniff' ); # defense from XSS
-    $self->res->body( $body );
-    $self->res;    
-}
-
-
-package Kossy::Request;
-
-use strict;
-use warnings;
-use parent qw/Plack::Request/;
-use Hash::MultiValue;
-use Encode;
-use Kossy::Validator;
-use Kossy::BodyParser;
-use Kossy::BodyParser::UrlEncoded;
-use Kossy::BodyParser::MultiPart;
-use Kossy::BodyParser::JSON;
-
-sub new {
-    my($class, $env, %opts) = @_;
-    Carp::croak(q{$env is required})
-        unless defined $env && ref($env) eq 'HASH';
-
-    bless {
-        %opts,
-        env => $env,
-    }, $class;
-}
-
-sub request_body_parser {
-    my $self = shift;
-    unless (exists $self->{request_body_parser}) {
-        $self->{request_body_parser} = $self->_build_request_body_parser();
-    }
-    return $self->{request_body_parser};
-}
-
-sub _build_request_body_parser {
-    my $self = shift;
-
-    my $parser = Kossy::BodyParser->new();
-    $parser->register(
-        'application/x-www-form-urlencoded',
-        'Kossy::BodyParser::UrlEncoded'
-    );
-    $parser->register(
-        'multipart/form-data',
-        'Kossy::BodyParser::MultiPart'
-    );
-    if ( $self->{parse_json_body} ) {
-            $parser->register(
-                'application/json',
-                'Kossy::BodyParser::JSON'
-            );
-    }
-    $parser;
-}
-
-sub _parse_request_body {
-    my $self = shift;
-    $self->request_body_parser->parse($self->env);
-}
-
-sub uploads {
-    my $self = shift;
-    unless ($self->env->{'kossy.request.upload_parameters'}) {
-        $self->_parse_request_body;
-    }
-    $self->env->{'plack.request.upload'} ||= 
-        Hash::MultiValue->new(@{$self->env->{'kossy.request.upload_parameters'}});
-}
-
-sub body_parameters {
-    my ($self) = @_;
-    $self->env->{'kossy.request.body'} ||= $self->_decode_parameters(@{$self->_body_parameters()});
-}
-
-sub query_parameters {
-    my ($self) = @_;
-    $self->env->{'kossy.request.query'} ||= $self->_decode_parameters(@{$self->_query_parameters()});
-}
-
-sub parameters {
-    my $self = shift;
-    $self->env->{'kossy.request.merged'} ||= do {
-        Hash::MultiValue->new(
-            $self->query_parameters->flatten,
-            $self->body_parameters->flatten,            
-        );
-    };
-}
-
-sub _decode_parameters {
-    my ($self, @flatten) = @_;
-    my @decoded;
-    while ( my ($k, $v) = splice @flatten, 0, 2 ) {
-        push @decoded, Encode::decode_utf8($k), Encode::decode_utf8($v);
-    }
-    return Hash::MultiValue->new(@decoded);
-}
-
-sub _body_parameters {
-    my $self = shift;
-    unless ($self->env->{'kossy.request.body_parameters'}) {
-        $self->_parse_request_body;
-    }
-    return $self->env->{'kossy.request.body_parameters'};    
-}
-
-sub _query_parameters {
-    my $self = shift;
-    unless ( $self->env->{'kossy.request.query_parameter'} ) {
-        $self->env->{'kossy.request.query_parameters'} = 
-            URL::Encode::url_params_flat($self->env->{'QUERY_STRING'});
-    }
-    return $self->env->{'kossy.request.query_parameters'};
-}
-
-sub body_parameters_raw {
-    my $self = shift;
-    unless ($self->env->{'plack.request.body'}) {
-        $self->env->{'plack.request.body'} = Hash::MultiValue->new(@{$self->_body_parameters});
-    }
-    return $self->env->{'plack.request.body'};
-}
-
-sub query_parameters_raw {
-    my $self = shift;
-    unless ($self->env->{'plack.request.query'}) {
-        $self->env->{'plack.request.query'} = Hash::MultiValue->new(@{$self->_query_parameters});
-    }
-    return $self->env->{'plack.request.query'};
-}
-
-sub parameters_raw {
-    my $self = shift;
-    $self->env->{'plack.request.merged'} ||= do {
-        Hash::MultiValue->new(
-            @{$self->_query_parameters},
-            @{$self->_body_parameters}
-        );
-    };
-}
-
-sub param_raw {
-    my $self = shift;
-
-    return keys %{ $self->parameters_raw } if @_ == 0;
-
-    my $key = shift;
-    return $self->parameters_raw->{$key} unless wantarray;
-    return $self->parameters_raw->get_all($key);
-}
-
-sub base {
-    my $self = shift;
-    $self->{_base} ||= {};
-    my $base = $self->_uri_base;
-    $self->{_base}->{$base} ||= $self->SUPER::base;
-    $self->{_base}->{$base}->clone;
-}
-
-sub uri_for {
-     my($self, $path, $args) = @_;
-     my $uri = $self->base;
-     my $base = $uri->path eq "/"
-              ? ""
-              : $uri->path;
-     $uri->path( $base . $path );
-     $uri->query_form(@$args) if $args;
-     $uri;
-}
-
-sub validator {
-    my ($self, $rule) = @_;
-    Kossy::Validator->check($self,$rule);
-}
-
-1;
-
-package Kossy::Response;
-
-use strict;
-use warnings;
-use parent qw/Plack::Response/;
-use Encode;
-use HTTP::Headers::Fast;
-use Cookie::Baker;
-
-sub headers {
-    my $self = shift;
-
-    if (@_) {
-        my $headers = shift;
-        if (ref $headers eq 'ARRAY') {
-            Carp::carp("Odd number of headers") if @$headers % 2 != 0;
-            $headers = HTTP::Headers::Fast->new(@$headers);
-        } elsif (ref $headers eq 'HASH') {
-            $headers = HTTP::Headers::Fast->new(%$headers);
-        }
-        return $self->{headers} = $headers;
-    } else {
-        return $self->{headers} ||= HTTP::Headers::Fast->new();
-    }
-}
-
-sub _body {
-    my $self = shift;
-    my $body = $self->body;
-       $body = [] unless defined $body;
-    if (!ref $body or Scalar::Util::blessed($body) && overload::Method($body, q("")) && !$body->can('getline')) {
-        return [ Encode::encode_utf8($body) ] if Encode::is_utf8($body);
-        return [ $body ];
-    } else {
-        return $body;
-    }
-}
-
-sub finalize {
-    my $self = shift;
-    Carp::croak "missing status" unless $self->status();
-
-    my @headers;
-    $self->headers->scan(sub{
-        my ($k,$v) = @_;
-        return if $k eq 'X-Frame-Options' || $k eq 'X-XSS-Protection';
-        $v =~ s/\015\012[\040|\011]+/chr(32)/ge; # replace LWS with a single SP
-        $v =~ s/\015|\012//g; # remove CR and LF since the char is invalid here
-        push @headers, $k, $v;
-    });
-
-    while (my($name, $val) = each %{$self->cookies}) {
-        my $cookie = bake_cookie($name, $val);
-        push @headers, 'Set-Cookie' => $cookie;
-    }
-
-    push @headers, 
-        'X-Frame-Options' => 'DENY',
-        'X-XSS-Protection' => 1;
-
-    return [
-        $self->status,
-        \@headers,
-        $self->_body,
-    ];
-}
-
 1;
 
 __END__
@@ -674,7 +290,7 @@ Kossy - Sinatra-ish Simple and Clear web application framework
 
 =head1 DESCRIPTION
 
-Kossy is Sinatra-ish Simple and Clear web application framework, which is based upon L<Plack>, L<Router::Simple>, L<Text::Xslate> and build-in Form-Validator. That's suitable for small application and rapid development.
+Kossy is Sinatra-ish Simple and Clear web application framework, which is based upon L<Plack>, L<Router::Boom>, L<Text::Xslate> and build-in Form-Validator. That's suitable for small application and rapid development.
 
 =head1 Kossy class
 
@@ -864,6 +480,67 @@ These methods are the accessor to raw values. 'raw' means the value is not decod
 
 This class is child class of Plack::Response
 
+=head1 CUSTOMIZE
+
+=over 4
+
+=item X-Frame-Options
+
+By default, Kossy outputs "X-Frame-Options: DENY". You can change this header 
+
+  get '/iframe' => sub {
+      my ($self, $c) = @_;
+      $c->res->header('X-Frame-Options','SAMEORIGIN');
+      # or remove from response header
+      # delete $c->res->headers->remove_header('X-Frame-Options');
+      ..
+  }
+
+(Default: DENY)
+
+=item kossy.request.parse_json_body
+
+If enabled, Kossy will decode json in the request body that has "application/json" content header
+
+  post '/api' => sub {
+      my ($self, $c) = @_;
+      $c->env->{'kossy.request.parse_json_body'} = 1;
+      my val = $c->req->param('foo'); # bar
+  }
+
+  # requrest
+  # $ua->requrest(
+  #     HTTP::Request->new(
+  #         "POST",
+  #         "http://example.com/api",
+  #         [ "Content-Type" => 'application/json', "Content-Length" => 13 ],
+  #         '{"foo":"bar"}'
+  #     )
+  # );
+
+=item $XSLATE_CACHE, $XSLATE_CACHE_DIR
+
+Change xslate's cache level and cache directory.
+
+  local $Kossy::XSLATE_CACHE = 2;
+  local $Kossy::XSLATE_CACHE_DIR = $dir;
+  my $app = MyApp::Web->psgi;
+
+By default, $XSLATE_CACHE is 1, $XSLATE_CACHE_DIR is undef. use Xslate's default.
+
+=item $SECURITY_HEADER
+
+If disabled, Kossy does not set X-Frame-Options and X-XSS-Protection. enabled by default.
+
+  local $Kossy::SECURITY_HEADER = 0;
+  my $app = MyApp::Web->psgi;
+
+Can not change $Kossy::SECURITY_HEADER in your WebApp. It's need to set at build time. 
+
+This is useful for the benchmark :-)
+
+=back
+
 =head1 AUTHOR
 
 Masahiro Nagano E<lt>kazeburo {at} gmail.comE<gt>
